@@ -22,6 +22,13 @@ INSERT INTO tbl VALUES(3, 'うえお');
 INSERT INTO tbl VALUES(10, 'ABC');
 INSERT INTO tbl VALUES(11, 'BCD');
 INSERT INTO tbl VALUES(12, 'CDE');
+INSERT INTO tbl VALUES(20, 'A B');
+INSERT INTO tbl VALUES(21, E'A \\ B');
+INSERT INTO tbl VALUES(22, 'A " B');
+INSERT INTO tbl VALUES(23, 'A * B');
+INSERT INTO tbl VALUES(24, 'A + B');
+INSERT INTO tbl VALUES(25, 'A % B');
+INSERT INTO tbl VALUES(26, 'A _ B');
 
 CREATE INDEX idx_senna ON tbl USING senna (t) WITH (initial_n_segments = 4);
 CREATE INDEX idx_like ON tbl USING senna (t like_ops) WITH (initial_n_segments = 2);
@@ -31,7 +38,9 @@ SELECT relname, reloptions FROM pg_class WHERE relname IN ('idx_senna', 'idx_lik
 SET enable_seqscan = on;
 SET enable_indexscan = off;
 SET enable_bitmapscan = off;
+SELECT * FROM tbl WHERE t %% '不在' ORDER BY id;
 SELECT * FROM tbl WHERE t %% 'いう' ORDER BY id;
+SELECT * FROM tbl WHERE t ~~% '%不在%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%いう%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% 'あ%う' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%BC%' ORDER BY id;
@@ -40,7 +49,9 @@ SELECT * FROM tbl WHERE t ~~% 'A%C' ORDER BY id;
 SET enable_seqscan = off;
 SET enable_indexscan = on;
 SET enable_bitmapscan = off;
+SELECT * FROM tbl WHERE t %% '不在' ORDER BY id;
 SELECT * FROM tbl WHERE t %% 'いう' ORDER BY id;
+SELECT * FROM tbl WHERE t ~~% '%不在%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%いう%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% 'あ%う' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%BC%' ORDER BY id;
@@ -49,7 +60,9 @@ SELECT * FROM tbl WHERE t ~~% 'A%C' ORDER BY id;
 SET enable_seqscan = off;
 SET enable_indexscan = off;
 SET enable_bitmapscan = on;
+SELECT * FROM tbl WHERE t %% '不在' ORDER BY id;
 SELECT * FROM tbl WHERE t %% 'いう' ORDER BY id;
+SELECT * FROM tbl WHERE t ~~% '%不在%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%いう%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% 'あ%う' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%BC%' ORDER BY id;
@@ -63,7 +76,9 @@ INSERT INTO tbl VALUES(15, 'abcde');
 SET enable_seqscan = on;
 SET enable_indexscan = off;
 SET enable_bitmapscan = off;
+SELECT * FROM tbl WHERE t %% '不在' ORDER BY id;
 SELECT * FROM tbl WHERE t %% 'いう' ORDER BY id;
+SELECT * FROM tbl WHERE t ~~% '%不在%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%いう%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% 'あ%う' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%BC%' ORDER BY id;
@@ -72,7 +87,9 @@ SELECT * FROM tbl WHERE t ~~% 'A%C' ORDER BY id;
 SET enable_seqscan = off;
 SET enable_indexscan = on;
 SET enable_bitmapscan = off;
+SELECT * FROM tbl WHERE t %% '不在' ORDER BY id;
 SELECT * FROM tbl WHERE t %% 'いう' ORDER BY id;
+SELECT * FROM tbl WHERE t ~~% '%不在%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%いう%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% 'あ%う' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%BC%' ORDER BY id;
@@ -81,7 +98,9 @@ SELECT * FROM tbl WHERE t ~~% 'A%C' ORDER BY id;
 SET enable_seqscan = off;
 SET enable_indexscan = off;
 SET enable_bitmapscan = on;
+SELECT * FROM tbl WHERE t %% '不在' ORDER BY id;
 SELECT * FROM tbl WHERE t %% 'いう' ORDER BY id;
+SELECT * FROM tbl WHERE t ~~% '%不在%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%いう%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% 'あ%う' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%BC%' ORDER BY id;
@@ -96,7 +115,9 @@ UPDATE tbl SET t = 'BCBC' WHERE id = 13;
 SET enable_seqscan = on;
 SET enable_indexscan = off;
 SET enable_bitmapscan = off;
+SELECT * FROM tbl WHERE t %% '不在' ORDER BY id;
 SELECT * FROM tbl WHERE t %% 'いう' ORDER BY id;
+SELECT * FROM tbl WHERE t ~~% '%不在%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%いう%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% 'あ%う' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%BC%' ORDER BY id;
@@ -106,7 +127,9 @@ SELECT * FROM tbl WHERE t ~~% 'A%C' ORDER BY id;
 SET enable_seqscan = off;
 SET enable_indexscan = on;
 SET enable_bitmapscan = off;
+SELECT * FROM tbl WHERE t %% '不在' ORDER BY id;
 SELECT * FROM tbl WHERE t %% 'いう' ORDER BY id;
+SELECT * FROM tbl WHERE t ~~% '%不在%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%いう%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% 'あ%う' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%BC%' ORDER BY id;
@@ -116,7 +139,9 @@ SELECT * FROM tbl WHERE t ~~% 'A%C' ORDER BY id;
 SET enable_seqscan = off;
 SET enable_indexscan = off;
 SET enable_bitmapscan = on;
+SELECT * FROM tbl WHERE t %% '不在' ORDER BY id;
 SELECT * FROM tbl WHERE t %% 'いう' ORDER BY id;
+SELECT * FROM tbl WHERE t ~~% '%不在%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%いう%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% 'あ%う' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%BC%' ORDER BY id;
@@ -128,7 +153,9 @@ VACUUM ANALYZE tbl;
 SET enable_seqscan = on;
 SET enable_indexscan = off;
 SET enable_bitmapscan = off;
+SELECT * FROM tbl WHERE t %% '不在' ORDER BY id;
 SELECT * FROM tbl WHERE t %% 'いう' ORDER BY id;
+SELECT * FROM tbl WHERE t ~~% '%不在%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%いう%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% 'あ%う' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%BC%' ORDER BY id;
@@ -137,7 +164,9 @@ SELECT * FROM tbl WHERE t ~~% 'A%C' ORDER BY id;
 SET enable_seqscan = off;
 SET enable_indexscan = on;
 SET enable_bitmapscan = off;
+SELECT * FROM tbl WHERE t %% '不在' ORDER BY id;
 SELECT * FROM tbl WHERE t %% 'いう' ORDER BY id;
+SELECT * FROM tbl WHERE t ~~% '%不在%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%いう%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% 'あ%う' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%BC%' ORDER BY id;
@@ -146,7 +175,9 @@ SELECT * FROM tbl WHERE t ~~% 'A%C' ORDER BY id;
 SET enable_seqscan = off;
 SET enable_indexscan = off;
 SET enable_bitmapscan = on;
+SELECT * FROM tbl WHERE t %% '不在' ORDER BY id;
 SELECT * FROM tbl WHERE t %% 'いう' ORDER BY id;
+SELECT * FROM tbl WHERE t ~~% '%不在%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%いう%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% 'あ%う' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%BC%' ORDER BY id;
@@ -154,44 +185,15 @@ SELECT * FROM tbl WHERE t ~~% 'A%C' ORDER BY id;
 
 DELETE FROM tbl WHERE id = 3;
 
-SET client_min_messages = error;
-VACUUM FULL tbl;
-SET client_min_messages = notice;
-
-SET enable_seqscan = on;
-SET enable_indexscan = off;
-SET enable_bitmapscan = off;
-SELECT * FROM tbl WHERE t %% 'いう' ORDER BY id;
-SELECT * FROM tbl WHERE t ~~% '%いう%' ORDER BY id;
-SELECT * FROM tbl WHERE t ~~% 'あ%う' ORDER BY id;
-SELECT * FROM tbl WHERE t ~~% '%BC%' ORDER BY id;
-SELECT * FROM tbl WHERE t ~~% 'A%C' ORDER BY id;
-
-SET enable_seqscan = off;
-SET enable_indexscan = on;
-SET enable_bitmapscan = off;
-SELECT * FROM tbl WHERE t %% 'いう' ORDER BY id;
-SELECT * FROM tbl WHERE t ~~% '%いう%' ORDER BY id;
-SELECT * FROM tbl WHERE t ~~% 'あ%う' ORDER BY id;
-SELECT * FROM tbl WHERE t ~~% '%BC%' ORDER BY id;
-SELECT * FROM tbl WHERE t ~~% 'A%C' ORDER BY id;
-
-SET enable_seqscan = off;
-SET enable_indexscan = off;
-SET enable_bitmapscan = on;
-SELECT * FROM tbl WHERE t %% 'いう' ORDER BY id;
-SELECT * FROM tbl WHERE t ~~% '%いう%' ORDER BY id;
-SELECT * FROM tbl WHERE t ~~% 'あ%う' ORDER BY id;
-SELECT * FROM tbl WHERE t ~~% '%BC%' ORDER BY id;
-SELECT * FROM tbl WHERE t ~~% 'A%C' ORDER BY id;
-
 SELECT senna.reindex_index('idx_senna');
 SELECT senna.reindex_index('idx_like');
 
 SET enable_seqscan = on;
 SET enable_indexscan = off;
 SET enable_bitmapscan = off;
+SELECT * FROM tbl WHERE t %% '不在' ORDER BY id;
 SELECT * FROM tbl WHERE t %% 'いう' ORDER BY id;
+SELECT * FROM tbl WHERE t ~~% '%不在%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%いう%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% 'あ%う' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%BC%' ORDER BY id;
@@ -200,7 +202,9 @@ SELECT * FROM tbl WHERE t ~~% 'A%C' ORDER BY id;
 SET enable_seqscan = off;
 SET enable_indexscan = on;
 SET enable_bitmapscan = off;
+SELECT * FROM tbl WHERE t %% '不在' ORDER BY id;
 SELECT * FROM tbl WHERE t %% 'いう' ORDER BY id;
+SELECT * FROM tbl WHERE t ~~% '%不在%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%いう%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% 'あ%う' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%BC%' ORDER BY id;
@@ -209,12 +213,39 @@ SELECT * FROM tbl WHERE t ~~% 'A%C' ORDER BY id;
 SET enable_seqscan = off;
 SET enable_indexscan = off;
 SET enable_bitmapscan = on;
+SELECT * FROM tbl WHERE t %% '不在' ORDER BY id;
 SELECT * FROM tbl WHERE t %% 'いう' ORDER BY id;
+SELECT * FROM tbl WHERE t ~~% '%不在%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%いう%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% 'あ%う' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% '%BC%' ORDER BY id;
 SELECT * FROM tbl WHERE t ~~% 'A%C' ORDER BY id;
 
+SELECT count(*) FROM tbl WHERE t ~~% '%いう%'
+UNION ALL
+SELECT count(*) FROM tbl WHERE t ~~% '%いう%';
+
+SELECT * FROM tbl WHERE t ~~% '%あい%' AND t ~~% '%お%';
+
+SELECT * FROM tbl WHERE t ~~% '%A B%';
+SELECT * FROM tbl WHERE t ~~% E'%A\\ B%';
+SELECT * FROM tbl WHERE t ~~% E'%A \\\\ B%';
+SELECT * FROM tbl WHERE t ~~% '%A " B%';
+SELECT * FROM tbl WHERE t ~~% '%A * B%';
+SELECT * FROM tbl WHERE t ~~% '%A + B%';
+SELECT * FROM tbl WHERE t ~~% E'%A \\% B%';
+SELECT * FROM tbl WHERE t ~~% E'%A \\_ B%';
+
+-- search for empty keys
+SELECT * FROM tbl WHERE t ~~% '';
+SELECT * FROM tbl WHERE t ~~% '%';
+SELECT * FROM tbl WHERE t ~~% '_';
+SELECT * FROM tbl WHERE t ~~% '%%';
+SELECT * FROM tbl WHERE t ~~% '%_%';
+SELECT * FROM tbl WHERE t ~~% '"';
+SELECT * FROM tbl WHERE t ~~% E'\\';
+SELECT * FROM tbl WHERE t ~~% NULL;
+
 SELECT senna.checkpoint();
 
 SELECT senna.drop_index('idx_senna');
