@@ -1,7 +1,7 @@
 /*
  * textsearch_senna.c
  *
- *	Copyright (c) 2009, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
+ *	Copyright (c) 2009-2010, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
  */
 #include "postgres.h"
 
@@ -26,7 +26,14 @@
 #include "nodes/parsenodes.h"
 #include "nodes/relation.h"
 #include "parser/parsetree.h"
+
+#if PG_VERSION_NUM >= 80300
 #include "postmaster/syslogger.h"
+#else
+/* 8.2 does not install postmaster headers !? */
+extern DLLIMPORT char *Log_directory;
+#endif
+
 #include "storage/fd.h"
 #include "storage/lmgr.h"
 #include "storage/bufmgr.h"
@@ -39,6 +46,8 @@
 #include "utils/selfuncs.h"
 #include "utils/syscache.h"
 
+#include "pgut/pgut-be.h"
+
 PG_MODULE_MAGIC;
 
 /*
@@ -146,30 +155,30 @@ PG_FUNCTION_INFO_V1(senna_costestimate);
 PG_FUNCTION_INFO_V1(senna_options);
 #endif
 
-extern void		_PG_init(void);
-extern void		_PG_fini(void);
-extern Datum pg_sync_file(PG_FUNCTION_ARGS);
-extern Datum senna_drop_index(PG_FUNCTION_ARGS);
-extern Datum senna_reindex_index(PG_FUNCTION_ARGS);
-extern Datum senna_contains(PG_FUNCTION_ARGS);
-extern Datum senna_contained(PG_FUNCTION_ARGS);
-extern Datum senna_restsel(PG_FUNCTION_ARGS);
-extern Datum senna_insert(PG_FUNCTION_ARGS);
-extern Datum senna_beginscan(PG_FUNCTION_ARGS);
-extern Datum senna_gettuple(PG_FUNCTION_ARGS);
+extern void PGUT_EXPORT _PG_init(void);
+extern void PGUT_EXPORT _PG_fini(void);
+extern Datum PGUT_EXPORT pg_sync_file(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_drop_index(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_reindex_index(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_contains(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_contained(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_restsel(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_insert(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_beginscan(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_gettuple(PG_FUNCTION_ARGS);
 #if PG_VERSION_NUM >= 80400
-extern Datum senna_getbitmap(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_getbitmap(PG_FUNCTION_ARGS);
 #else
-extern Datum senna_getmulti(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_getmulti(PG_FUNCTION_ARGS);
 #endif
-extern Datum senna_rescan(PG_FUNCTION_ARGS);
-extern Datum senna_endscan(PG_FUNCTION_ARGS);
-extern Datum senna_build(PG_FUNCTION_ARGS);
-extern Datum senna_bulkdelete(PG_FUNCTION_ARGS);
-extern Datum senna_vacuumcleanup(PG_FUNCTION_ARGS);
-extern Datum senna_costestimate(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_rescan(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_endscan(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_build(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_bulkdelete(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_vacuumcleanup(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_costestimate(PG_FUNCTION_ARGS);
 #if PG_VERSION_NUM >= 80200
-extern Datum senna_options(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT senna_options(PG_FUNCTION_ARGS);
 #endif
 
 static char *SennaPath(const RelFileNode *rnode);
@@ -180,58 +189,20 @@ static bool SennaRemove(const RelFileNode *rnode);
 static sen_query *SennaQuery(const char *key, size_t len);
 static sen_records *SennaRecordsOpen(Relation indexRelation, const char *str, int len, sen_index **idx);
 static void SennaRecordsClose(sen_records *records);
-static void SennaLock(Relation index, LOCKMODE lockmode);
-static void SennaUnlock(Relation index, LOCKMODE lockmode);
-static void SennaScanOpen(SennaScanDesc *desc, IndexScanDesc scan, StrategyNumber strategy, Datum datum);
-static bool SennaScanNext(SennaScanDesc *desc, ItemPointer ctid);
+static void SennaLock(Relation index);
+static void SennaUnlock(Relation index);
+static bool SennaScanNext(SennaScanDesc *desc, IndexScanDesc scan, ItemPointer ctid);
 static void SennaScanClose(SennaScanDesc *desc);
 static void SennaInsert(sen_index *index, ItemPointer ctid, Datum value);
 static void SennaDelete(sen_index *index, ItemPointer key, text *value);
 static void SennaBuildCallback(Relation index, HeapTuple htup, Datum *values, bool *isnull, bool tupleIsAlive, void *context);
-static size_t SennaIndexSize(sen_index *index);
+static uint64 SennaIndexSize(sen_index *index);
 static IndexBulkDeleteResult *SennaBulkDeleteResult(sen_index *index);
 static void SennaXactCallback(XactEvent event, void *arg);
 static void SennaCacheRelCallback(Datum arg, Oid relid);
 
-#if PG_VERSION_NUM < 80300
-
-#define PG_GETARG_TEXT_PP(n)		PG_GETARG_TEXT_P(n)
-#define VARSIZE_ANY_EXHDR(v)		(VARSIZE(v) - VARHDRSZ)
-#define VARDATA_ANY(v)				VARDATA(v)
-#define SET_VARSIZE(v, sz)			(VARATT_SIZEP(v) = (sz))
-#define pg_detoast_datum_packed(v)	pg_detoast_datum(v)
-#define DatumGetTextPP(v)			DatumGetTextP(v)
-#define ItemIdIsNormal(v)			ItemIdIsUsed(v)
-#define IndexBuildHeapScan(heap, index, info, sync, callback, state) \
-	IndexBuildHeapScan((heap), (index), (info), (callback), (state))
-#define planner_rt_fetch(rti, root) \
-	 rt_fetch(rti, (root)->parse->rtable)
-
-#endif
-
 #if PG_VERSION_NUM < 80400
 
-#define MAIN_FORKNUM	0
-#define relpath(rnode, forknum)		relpath((rnode))
-
-static char *
-text_to_cstring(const text *t)
-{
-	/* must cast away the const, unfortunately */
-	text	   *tunpacked = pg_detoast_datum_packed((struct varlena *) t);
-	int			len = VARSIZE_ANY_EXHDR(tunpacked);
-	char	   *result;
-
-	result = (char *) palloc(len + 1);
-	memcpy(result, VARDATA_ANY(tunpacked), len);
-	result[len] = '\0';
-
-	if (tunpacked != t)
-		pfree(tunpacked);
-	
-	return result;
-}
-
 static void
 RemoveRelations(DropStmt *drop)
 {
@@ -419,7 +390,7 @@ senna_reindex_index(PG_FUNCTION_ARGS)
 	rnode = index->rd_node;
 	index_close(index, NoLock);
 
-	reindex_index(relid);
+	reindex_index(relid, false);
 
 	SennaRemove(&rnode);
 
@@ -518,7 +489,7 @@ senna_restsel(PG_FUNCTION_ARGS)
 				text		   *query;
 				sen_index	   *index;
 				sen_records	   *records;
-				int				nhits;
+				unsigned int	nhits;
 				unsigned int	nkeys;
 
 				/*
@@ -579,9 +550,9 @@ senna_insert(PG_FUNCTION_ARGS)
 
 	i = SennaIndexOpen(index, 0);
 
-	SennaLock(index, ExclusiveLock);
+	SennaLock(index);
 	SennaInsert(i, ctid, values[0]);
-	SennaUnlock(index, ExclusiveLock);
+	SennaUnlock(index);
 
 	PG_RETURN_BOOL(true);
 }
@@ -591,7 +562,7 @@ senna_beginscan(PG_FUNCTION_ARGS)
 {
 	Relation		index = (Relation) PG_GETARG_POINTER(0);
 	int				keysz = PG_GETARG_INT32(1);
-	ScanKey			scankey = (ScanKey) PG_GETARG_POINTER(2);
+	ScanKey			key = (ScanKey) PG_GETARG_POINTER(2);
 	IndexScanDesc	scan;
 
 	if (keysz < 1)
@@ -599,7 +570,7 @@ senna_beginscan(PG_FUNCTION_ARGS)
 				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
 				 errmsg("senna: do not support whole-index scans")));
 
-	scan = RelationGetIndexScan(index, keysz, scankey);
+	scan = RelationGetIndexScan(index, keysz, key);
 
 	PG_RETURN_POINTER(scan);
 }
@@ -652,7 +623,7 @@ senna_gettuple(PG_FUNCTION_ARGS)
 	IndexScanDesc	scan = (IndexScanDesc) PG_GETARG_POINTER(0);
 	ScanDirection	dir = (ScanDirection) PG_GETARG_INT32(1);
 
-	SennaScanDesc  *desc = (SennaScanDesc *) scan->opaque;
+	SennaScanDesc *desc = (SennaScanDesc *) scan->opaque;
 	ItemPointerData ctid;
 
 	if (dir != ForwardScanDirection)
@@ -664,9 +635,9 @@ senna_gettuple(PG_FUNCTION_ARGS)
 		text	   *value = extract_value(&scan->xs_ctup,
 								scan->heapRelation, scan->indexRelation);
 
-		SennaLock(scan->indexRelation, ExclusiveLock);
+		SennaLock(scan->indexRelation);
 		SennaDelete(index, &scan->xs_ctup.t_self, value);
-		SennaUnlock(scan->indexRelation, ExclusiveLock);
+		SennaUnlock(scan->indexRelation);
 	}
 
 #if NOT_USED
@@ -682,13 +653,14 @@ senna_gettuple(PG_FUNCTION_ARGS)
 	}
 #endif
 
-#if PG_VERSION_NUM >= 80400
-	scan->xs_recheck = desc->recheck;
-#endif
-	if (!SennaScanNext(desc, &ctid))
+	if (!SennaScanNext(desc, scan, &ctid))
 		PG_RETURN_BOOL(false);
 
 	scan->xs_ctup.t_self = ctid;
+#if PG_VERSION_NUM >= 80400
+	scan->xs_recheck = desc->recheck;
+#endif
+
 	PG_RETURN_BOOL(true);
 }
 
@@ -700,11 +672,11 @@ senna_getbitmap(PG_FUNCTION_ARGS)
 	IndexScanDesc	scan = (IndexScanDesc) PG_GETARG_POINTER(0);
 	TIDBitmap	   *tbm = (TIDBitmap *) PG_GETARG_POINTER(1);
 
-	SennaScanDesc  *desc = (SennaScanDesc *) scan->opaque;
+	SennaScanDesc *desc = (SennaScanDesc *) scan->opaque;
 	int64			ntids;
 	ItemPointerData ctid;
 
-	for (ntids = 0; SennaScanNext(desc, &ctid); ntids++)
+	for (ntids = 0; SennaScanNext(desc, scan, &ctid); ntids++)
 	{
 		CHECK_FOR_INTERRUPTS();
 		tbm_add_tuples(tbm, &ctid, 1, desc->recheck);
@@ -723,7 +695,7 @@ senna_getmulti(PG_FUNCTION_ARGS)
 	int32		max_tids = PG_GETARG_INT32(2);
 	int32	   *returned_tids = (int32 *) PG_GETARG_POINTER(3);
 
-	SennaScanDesc  *desc = (SennaScanDesc *) scan->opaque;
+	SennaScanDesc *desc = (SennaScanDesc *) scan->opaque;
 	int64			ntids;
 	ItemPointerData ctid;
 
@@ -731,7 +703,7 @@ senna_getmulti(PG_FUNCTION_ARGS)
 		PG_RETURN_BOOL(true);
 
 	for (ntids = 0;
-		 ntids < max_tids && SennaScanNext(desc, &ctid);
+		 ntids < max_tids && SennaScanNext(desc, scan, &ctid);
 		 ntids++)
 	{
 		/* Save tuple ID, and continue scanning */
@@ -748,9 +720,9 @@ Datum
 senna_rescan(PG_FUNCTION_ARGS)
 {
 	IndexScanDesc	scan = (IndexScanDesc) PG_GETARG_POINTER(0);
-	ScanKey			scankey = (ScanKey) PG_GETARG_POINTER(1);
+	ScanKey			key = (ScanKey) PG_GETARG_POINTER(1);
 
-	SennaScanDesc  *desc = (SennaScanDesc  *) scan->opaque;
+	SennaScanDesc *desc = (SennaScanDesc *) scan->opaque;
 
 	if (desc == NULL)
 	{
@@ -763,8 +735,8 @@ senna_rescan(PG_FUNCTION_ARGS)
 	else
 		SennaScanClose(desc);
 
-	if (scankey && scan->numberOfKeys > 0)
-		SennaScanOpen(desc, scan, scankey[0].sk_strategy, scankey[0].sk_argument);
+	if (key && scan->numberOfKeys > 0)
+		memmove(scan->keyData, key, scan->numberOfKeys * sizeof(ScanKeyData));
 
 	PG_RETURN_VOID();
 }
@@ -774,7 +746,7 @@ senna_endscan(PG_FUNCTION_ARGS)
 {
 	IndexScanDesc	scan = (IndexScanDesc) PG_GETARG_POINTER(0);
 
-	SennaScanDesc  *desc = (SennaScanDesc *) scan->opaque;
+	SennaScanDesc *desc = (SennaScanDesc *) scan->opaque;
 
 	if (desc != NULL)
 	{
@@ -837,10 +809,9 @@ senna_build(PG_FUNCTION_ARGS)
 	PG_TRY();
 	{
 		result = (IndexBuildResult *) palloc(sizeof(IndexBuildResult));
-		SennaLock(index, ExclusiveLock);
+		SennaLock(index);
 		result->heap_tuples = IndexBuildHeapScan(heap, index, indexInfo, true, SennaBuildCallback, &state);
-
-		SennaUnlock(index, ExclusiveLock);
+		SennaUnlock(index);
 		result->index_tuples = sen_sym_size(state.index->keys);
 	}
 	PG_CATCH();
@@ -933,9 +904,9 @@ senna_bulkdelete(PG_FUNCTION_ARGS)
 				value = extract_value(&tuple, heap, info->index);
 			}
 #endif
-			SennaLock(info->index, ExclusiveLock);
+			SennaLock(info->index);
 			SennaDelete(index, &ctid, value);
-			SennaUnlock(info->index, ExclusiveLock);
+			SennaUnlock(info->index);
 
 #ifdef VACUUM_USES_HEAP
 			ReleaseBuffer(buffer);
@@ -1218,7 +1189,7 @@ SennaRemove(const RelFileNode *rnode)
 }
 
 static sen_query *
-SennaQuery(const char *sql, size_t len)
+SennaQuery(const char *str, size_t len)
 {
 	int					i;
 	SennaQueryCache    *cache;
@@ -1232,7 +1203,7 @@ SennaQuery(const char *sql, size_t len)
 
 		if (cache->query == NULL)
 			break;
-		if (strncmp(cache->key, sql, len) == 0)
+		if (strncmp(cache->key, str, len) == 0)
 			return cache->query;
 	}
 
@@ -1242,8 +1213,8 @@ SennaQuery(const char *sql, size_t len)
 		ereport(ERROR,
 				(errcode(ERRCODE_OUT_OF_MEMORY),
 				 errmsg("out of memory")));
-	strlcpy(key, sql, len + 1);
-	query = sen_query_open(sql, len, sen_sel_or, SENNA_MAX_N_EXPR, SennaEncoding());
+	strlcpy(key, str, len + 1);
+	query = sen_query_open(str, len, sen_sel_or, SENNA_MAX_N_EXPR, SennaEncoding());
 	if (query == NULL)
 		elog(ERROR, "senna: sen_query_open()");
 
@@ -1276,6 +1247,13 @@ SennaRecordsOpen(Relation indexRelation, const char *str, int len, sen_index **i
 	ListCell		   *cell;
 	SennaRecordsCache	*cache;
 
+	if (len == 0)
+		ereport(ERROR,
+			(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+			 errmsg("scan key must not be empty for senna index: %s.%s",
+					get_namespace_name(RelationGetNamespace(indexRelation)),
+					 RelationGetRelationName(indexRelation))));
+
 	query = SennaQuery(str, len);
 
 	/* search records cache first */
@@ -1289,6 +1267,7 @@ SennaRecordsOpen(Relation indexRelation, const char *str, int len, sen_index **i
 			records = cache->records;
 			if (idx)
 				*idx = SennaIndexOpen(indexRelation, 0);
+			sen_records_rewind(records);
 			return records;
 		}
 	}
@@ -1309,13 +1288,8 @@ SennaRecordsOpen(Relation indexRelation, const char *str, int len, sen_index **i
 
 	index = SennaIndexOpen(indexRelation, 0);
 
-	/*
-	 * Use sen_query_exec() instead of sen_index_select() in order to
-	 * minimize locking duration. Compile queries before locking.
-	 */
-	SennaLock(indexRelation, ShareLock);
+	/* Senna doesn't require any locks in reads. */
 	rc = sen_query_exec(index, query, records, sen_sel_or);
-	SennaUnlock(indexRelation, ShareLock);
 
 	if (rc != sen_success)
 		elog(ERROR, "senna: sen_query_exec() : code=%d", rc);
@@ -1354,28 +1328,33 @@ SennaRecordsClose(sen_records *records)
 }
 
 static void
-SennaLock(Relation index, LOCKMODE lockmode)
+SennaLock(Relation index)
 {
 	const RelFileNode *rnode = &index->rd_node;
 	LockDatabaseObject(rnode->spcNode,
 					   rnode->dbNode,
 					   rnode->relNode,
-					   lockmode);
+					   ExclusiveLock);
 }
 
 static void
-SennaUnlock(Relation index, LOCKMODE lockmode)
+SennaUnlock(Relation index)
 {
 	const RelFileNode *rnode = &index->rd_node;
 	UnlockDatabaseObject(rnode->spcNode,
 						 rnode->dbNode,
 						 rnode->relNode,
-						 lockmode);
+						 ExclusiveLock);
 }
 
-#define IS_LIKE_WILDCARD(c)		((c) == '%' || (c) == '_')
-#define LIKE_PRAGMA				"*D+ "
-#define LIKE_PRAGMA_LEN			4
+/*
+ * Note that senna cannot search " in words because the double quote is a
+ * special marker to separate words. I cannot find any descriptions in
+ * senna documentation to escape the character.
+ */
+#define LIKE_WILDCARD			"%_"
+#define WORD_SEP_CHARS			LIKE_WILDCARD "\""
+#define NON_SIMPLE_CHARS		WORD_SEP_CHARS " \\\t\v\r\n"
 
 /*
  * is_simple_like - true iff str matches %foo%.
@@ -1387,95 +1366,157 @@ is_simple_like(const char *str, int len)
 {
 	int		i;
 
-	if (len < 2 || str[0] != '%' || str[len - 1] != '%')
+	if (len <= 2 || str[0] != '%' || str[len - 1] != '%')
 		return false;
 
 	for (i = 1; i < len - 1; i++)
 	{
-		if (IS_LIKE_WILDCARD(str[i]) ||
-			isspace((unsigned char) str[i]) ||
-			str[i] == '\\')
+		if (strchr(NON_SIMPLE_CHARS, str[i]))
 			return false;
 	}
 	return true;
 }
 
+/*
+ * Concat multiple keys into one senna query.
+ * For senna index, "col %% 'e1' AND col %% 'e2'" will be 'e1+e2'.
+ * For like index, "col LIKE '%w1%' AND col LIKE '%w2%'" will be '"w1"+"w2"'.
+ */
 static void
-SennaScanOpen(SennaScanDesc *desc, IndexScanDesc scan, StrategyNumber strategy, Datum datum)
+SennaScanOpen(SennaScanDesc *desc,
+			  Relation indexRelation,
+			  int nkeys,
+			  const ScanKeyData keys[/*nkeys*/])
 {
+	StringInfoData	buf;
 	text		   *query;
 	const char	   *str;
 	int				len;
+	int				n;
 
-	query = DatumGetTextPP(datum);
-	str = VARDATA_ANY(query);
-	len = VARSIZE_ANY_EXHDR(query);
-
-	/* queyr is LIKE. convert LIKE query to senna query */
-	if (STRATEGY_IS_LIKE(strategy))
+	/* fast path for scanning senna index with just one key. */
+	if (nkeys == 1 && !STRATEGY_IS_LIKE(keys[0].sk_strategy))
 	{
-		if (is_simple_like(str, len))
+		query = DatumGetTextPP(keys[0].sk_argument);
+		str = VARDATA_ANY(query);
+		len = VARSIZE_ANY_EXHDR(query);
+
+		desc->records = SennaRecordsOpen(indexRelation, str, len, NULL);
+		return;
+	}
+
+	/* scan with multiple keys or LIKE strategy */
+	initStringInfo(&buf);
+	for (n = 0; n < nkeys; n++)
+	{
+		query = DatumGetTextPP(keys[n].sk_argument);
+		str = VARDATA_ANY(query);
+		len = VARSIZE_ANY_EXHDR(query);
+
+		if (STRATEGY_IS_LIKE(keys[n].sk_strategy))
 		{
-			/* don't need recheck if simple %foo% query */
-			str += 1;
-			len -= 2;
+			/* convert LIKE query to senna query */
+			if (is_simple_like(str, len))
+			{
+				/* %foo% is converted to "foo" */
+				if (buf.len > 0)
+					appendStringInfoChar(&buf, '+');
+				appendStringInfoChar(&buf, '"');
+				appendBinaryStringInfo(&buf, str + 1, len - 2);
+				appendStringInfoChar(&buf, '"');
+
+				/* don't need recheck if simple %foo% query */
+			}
+			else
+			{
+				int			i;
+				bool		quoted = false;
+
+				/* LIKE query needs recheck */
+				desc->recheck = true;
+
+				/* skip wildcard at head */
+				i = (int) strspn(str, LIKE_WILDCARD);
+				str += i;
+				len -= i;
+
+				/* convert LIKE pattern into senna query */
+				for (i = 0; i < len; i++)
+				{
+					if (strchr(WORD_SEP_CHARS, str[i]))
+					{
+						/* close the previous quote if open */
+						if (quoted)
+						{
+							appendStringInfoChar(&buf, '"');
+							quoted = false;
+						}
+					}
+					else
+					{
+						if (str[i] == '\\')
+						{
+							if (i >= len - 1)
+								ereport(ERROR,
+									(errcode(ERRCODE_INVALID_ESCAPE_SEQUENCE),
+									 errmsg("LIKE pattern must not end with escape character")));
+							/* skip the escape sequence and copy the next char. */
+							i++;
+						}
+
+						if (!quoted)
+						{
+							if (buf.len > 0)
+								appendStringInfoChar(&buf, '+');
+							appendStringInfoChar(&buf, '"');
+							quoted = true;
+						}
+
+						/* escape backslash in senna query */
+						if (str[i] == '\\')
+							appendStringInfoChar(&buf, '\\');
+						appendStringInfoChar(&buf, str[i]);
+					}
+				}
+
+				if (quoted)
+					appendStringInfoChar(&buf, '"');
+			}
 		}
 		else
 		{
-			char	   *buffer;
-			char	   *s;
-			int			i;
-			int			wilds;
-
-			/* LIKE query needs recheck */
-			desc->recheck = true;
-
-			buffer = malloc(len + LIKE_PRAGMA_LEN);
-			memcpy(buffer, LIKE_PRAGMA, LIKE_PRAGMA_LEN);
-
-			/* skip wildcard at head */
-			while (IS_LIKE_WILDCARD(*str))
-			{
-				str++;
-				len--;
-			}
-
-			/* copy query. wildcards are converted to '+' */
-			for (i = 0, s = buffer + LIKE_PRAGMA_LEN, wilds = 0; i < len; i++)
-			{
-				if (str[i] == '\\')
-				{
-					if (i >= len - 1)
-						ereport(ERROR,
-							(errcode(ERRCODE_INVALID_ESCAPE_SEQUENCE),
-							 errmsg("LIKE pattern must not end with escape character")));
-					/* skip the escape sequence and copy the next char. */
-					*s++ = str[++i];
-					wilds = 0;
-				}
-				else if (IS_LIKE_WILDCARD(str[i]))
-				{
-					if (wilds == 0)
-						*s++ = ' ';
-					wilds++;
-				}
-				else
-				{
-					*s++ = str[i];
-					wilds = 0;
-				}
-			}
-			str = buffer;
-			len = s - buffer - wilds;
+			if (buf.len > 0)
+				appendStringInfoChar(&buf, '+');
+			/* just copy for senna query */
+			appendBinaryStringInfo(&buf, str, len);
 		}
 	}
 
-	desc->records = SennaRecordsOpen(scan->indexRelation, str, len, NULL);
+	desc->records = SennaRecordsOpen(indexRelation, buf.data, buf.len, NULL);
+
+	pfree(buf.data);
 }
 
+/*
+ * Scan senna index or return the next tuple.
+ */
 static bool
-SennaScanNext(SennaScanDesc *desc, ItemPointer ctid)
+SennaScanNext(SennaScanDesc *desc, IndexScanDesc scan, ItemPointer ctid)
 {
+	if (desc->records == NULL)
+	{
+		int					i,
+							nkeys = scan->numberOfKeys;
+
+		/* NULL key is not supported */
+		for (i = 0; i < nkeys; i++)
+			if (scan->keyData[i].sk_flags & SK_ISNULL)
+				return false;
+
+		SennaScanOpen(desc, scan->indexRelation, nkeys, scan->keyData);
+	}
+
+	Assert(desc->records != NULL);
 	return sen_records_next(desc->records, ctid, SizeOfIptrData, NULL) == SizeOfIptrData;
 }
 
@@ -1539,7 +1580,7 @@ SennaDelete(sen_index *index, ItemPointer ctid, text *value)
 /*
  * Total index size in bytes.
  */
-static size_t
+static uint64
 SennaIndexSize(sen_index *index)
 {
 	unsigned file_size_keys;
